/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.email.security ;


import com.funambol.email.admin.EmailConnectorConfig;
import com.funambol.email.exception.EmailAccessException;
import com.funambol.email.exception.EmailConfigException;
import com.funambol.email.transport.IMailServerWrapper;
import com.funambol.email.transport.MailServerWrapperFactory;
import com.funambol.email.util.Def;
import com.funambol.framework.core.Authentication;
import com.funambol.framework.core.Cred;
import com.funambol.framework.filter.WhereClause;
import com.funambol.framework.security.Officer;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.tools.Base64;
import com.funambol.server.admin.UserManager;
import com.funambol.server.config.Configuration;

import java.security.Principal;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * This is an implementation of the <i>Officier</i> interface.
 * It always authenticates and authorizes users and resource accesses.
 *
 * @version $Id: EmailOfficer.java,v 1.14 2006/08/08 09:32:29 gibi Exp $
 */
public class EmailOfficer implements Officer, java.io.Serializable {

    // ---------------------------------------------------------- Protected data

    protected Logger log = Logger.getLogger(Def.LOGGER_NAME);

    protected PersistentStore ps = null;

    protected UserManager userManager = null;

    private static final String DEFAULT_USER_ROLE = "sync_user";

    // ------------------------------------------------------------- Public data

    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for incorrect login/password?
     */
    private boolean loginFailed = false;

    /**
     * @return boolean
     */
    public boolean isLoginFailed() {
        return loginFailed;
    }

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Which type of authetication does impose the server?
     */
    private String clientAuth = Cred.AUTH_TYPE_BASIC;

    /**
     * @return String
     */
    public String getClientAuth() {
        return this.clientAuth;
    }

    /**
     * @param clientAuth String
     */
    public void setClientAuth(String clientAuth) {
        this.clientAuth = clientAuth;
    }

    /**
     * Which type of authetication use the server to send his credential?
     */
    private String serverAuth = Cred.AUTH_NONE;

    /**
     * @return String
     */
    public String getServerAuth() {
        return this.serverAuth;
    }


    /**
     * @param serverAuth String
     */
    public void setServerAuth(String serverAuth) {
        this.serverAuth = serverAuth;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Authenticates a credential.
     *
     * @param credential the credential to be authenticated
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        Configuration config = Configuration.getConfiguration();
        userManager = config.getUserManager();

        ps = config.getStore();

        String type = credential.getType();

        if ((Cred.AUTH_TYPE_BASIC).equals(type)) {
            return authenticateBasicCredential(credential);
        }
        return false;
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource  the name (or the identifier) of the resource to be authorized
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /**
     * Un-authenticates a credential.
     * Do nothing. In the current implementation, the authentication is
     * discarde as soon as the LoginContext is garbage collected.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
    }

    /**
     * @return boolean
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }

    // --------------------------------------------------------- Private methods

    /**
     * @param credential Cred
     * @return boolean
     */
    private boolean authenticateBasicCredential(Cred credential) {

        String username, password;

        Authentication auth = credential.getAuthentication();
        String deviceId = auth.getDeviceId();
        String userpwd = new String(Base64.decode(auth.getData()));

        int p = userpwd.indexOf(':');

        if (p == -1) {
            username = userpwd;
            password = "";
        } else {
            username = (p > 0) ? userpwd.substring(0, p) : "";
            password = (p == (userpwd.length() - 1)) ? "" : userpwd.substring(p + 1);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Username: " + username);
        }

        //
        // Verify that exist the principal for the
        // specify deviceId and username
        //

        //
        // Verify these credentials on Mail Server
        //
        if (!checkMailCredentials(username, password)) {
            return false;
        }

        try {
            if (existsUser(username)) {
                if (log.isLoggable(Level.INFO)) {
                    log.info("User with '" + username + "' exists.");
                }
            } else {
                insertUser(username, " ");
                if (log.isLoggable(Level.INFO)) {
                    log.info("User '" + username + "' created.");
                }
            }
        }
        catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error inserting a new user: " + e.getMessage());
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, "authenticateBasicCredential", e);
            }
            return false;
        }

        try {
            if (!existsPrincipal(username, deviceId)) {
                credential.getAuthentication().setPrincipalId(insertPrincipal(username, deviceId));
            }
        }
        catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error inserting a new principal: " + e.getMessage());
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, "authenticateBasicCredential", e);
            }
            return false;
        }

        return true;
    }

    /**
     * Checks if the given login and password are valid
     * credentials on Mail Server
     *
     * @param login    String
     * @param password String
     * @return boolean
     */
    private boolean checkMailCredentials(String login,
                                         String password) {
        String host = "";
        String protocol = "pop3";
        int port = 110;
        boolean isSSL = false;
        String keystorePath = "";
        String keystorePassphrase = "";

        try {
            host = EmailConnectorConfig.getConfigInstance().getHost();
            protocol = EmailConnectorConfig.getConfigInstance().getProtocol();
            port = EmailConnectorConfig.getConfigInstance().getPort();
            isSSL = EmailConnectorConfig.getConfigInstance().getSsl();
            keystorePath = EmailConnectorConfig.getConfigInstance().getKeystorePath();
            keystorePassphrase = EmailConnectorConfig.getConfigInstance().getKeystorePassPhrase();
        }
        catch (EmailConfigException e) {
            log.severe("Error Getting Connector Parameters:: " + e);
            log.throwing(getClass().getName(), "checkMailCredentials", e);
        }

        IMailServerWrapper mswf =
                MailServerWrapperFactory.getMailServerWrapper(protocol);


        boolean isconnected = false;
        boolean error = false;
        try {
            if (mswf == null) {
                throw new EmailAccessException("Error setting Officer Protocol");
            }
            mswf.checkUser(host,
                    "" + port,
                    login,
                    password,
                    isSSL,
                    keystorePath,
                    keystorePassphrase);

        }
        catch (EmailAccessException dae) {
            log.severe("Error getting connection: " + dae);
            log.throwing(getClass().getName(), "checkMailCredentials", dae);
            error = true;
        }

        if (!error) {
            isconnected = true;
        }

        return isconnected;
    }

    /**
     * Checks if there is an user with the given username
     *
     * @param userName the username
     * @return true if the user exists, false otherwise
     */
    private boolean existsUser(String userName)
            throws PersistentStoreException {
        Sync4jUser[] users;
        WhereClause wc;
        String value[] = new String[]{userName};
        wc = new WhereClause("username", value, WhereClause.OPT_EQ, true);
        users = userManager.getUsers(wc);
        return users.length > 0;
    }

    /**
     * Insert a new user with the given username and password
     *
     * @param userName the username
     * @param password the password
     * @throws PersistentStoreException if an error occurs
     */
    private void insertUser(String userName, String password)
            throws PersistentStoreException {
        Sync4jUser user = new Sync4jUser();
        user.setUsername(userName);
        user.setPassword(password);
        user.setRoles(new String[]{DEFAULT_USER_ROLE});
        userManager.insertUser(user);
    }

    /**
     * Verify if there is a principal with the given userName and deviceId in store
     *
     * @param userName the username
     * @param deviceId the device's id
     * @throws PersistentStoreException if an error occurs
     */
    private boolean existsPrincipal(String userName, String deviceId)
            throws PersistentStoreException {
        Principal principal;
        try {
            principal = Sync4jPrincipal.createPrincipal(userName, deviceId);
            ps.read(principal);
            return true;
        }
        catch (NotFoundException nfe) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Principal for " + userName + ":" + deviceId + " not found");
            }
            return false;
        }

    }

    /**
     * Insert a principal
     *
     * @param userName the username
     * @param deviceId the device's id
     * @return the id of the principal
     * @throws PersistentStoreException if an error occurs
     */
    private long insertPrincipal(String userName, String deviceId)
            throws PersistentStoreException {
        Sync4jPrincipal principal = Sync4jPrincipal.createPrincipal(userName, deviceId);
        try {
            ps.store(principal);
        }
        catch (PersistentStoreException ex) {
            ex.printStackTrace();
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating new principal: " + ex);
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, "authenticateBasicCredential", ex);
            }
            throw ex;
        }
        return principal.getId();
    }
}
